# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import jax
import jax.numpy as jnp
from numpyro.infer import Predictive
from typing import Dict

from twinify.dpvi import ModelFunction, GuideFunction

def sample_synthetic_data(
        model: ModelFunction,
        guide: GuideFunction,
        posterior_params: Dict[str, jnp.ndarray],
        sampling_rng: jnp.ndarray,
        num_parameter_samples: int,
        num_record_samples_per_parameter_sample: int
    ) -> Dict[str, jnp.ndarray]:
    """
    Args:
        model: Model function.
        guide: Guide // variational approximation function.
        posterior_params: Dictionary of learned parameters for the guide/model.
        sampling_rng: JAX RNG key for sampling.
        num_parameter_samples: How many different parameter sets to sample from the guide.
            For each of these a synthetic dataset will be sampled.
        num_record_samples_per_parameter_sample: The number of synthetic data points sampled for each parameter set sample.
    Returns:
        A dictionary of np.arrays of synthetic data samples for each sample site in the model.
        Each of these has shape (num_parameter_samples, num_synthetic_records_per_parameter_sample, *sample_site_shape)
    """

    @jax.jit
    def sample_from_ppd(rng_key):
        """ Samples a single parameter vector and
            num_record_samples_per_parameter_sample based on it.
        """
        parameter_sampling_rng, record_sampling_rng = jax.random.split(rng_key)

        # sample single parameter vector
        posterior_sampler = Predictive(
            guide, params=posterior_params, num_samples=1
        )
        posterior_samples = posterior_sampler(parameter_sampling_rng)
        # models always add a superfluous batch dimensions, squeeze it
        posterior_samples = {k: v.squeeze(0) for k,v in posterior_samples.items()}

        # sample num_record_samples_per_parameter_sample data samples
        ppd_sampler = Predictive(model, posterior_samples, batch_ndims=0)
        per_sample_rngs = jax.random.split(
            record_sampling_rng, num_record_samples_per_parameter_sample
        )
        ppd_samples = jax.vmap(ppd_sampler)(per_sample_rngs)
        # models always add a superfluous batch dimensions, squeeze it
        ppd_samples = {k: v.squeeze(1) for k, v in ppd_samples.items()}

        return ppd_samples

    per_parameter_rngs = jax.random.split(sampling_rng, num_parameter_samples)
    ppd_samples = jax.vmap(sample_from_ppd)(per_parameter_rngs)

    return {site: np.asarray(value) for site, value in ppd_samples.items()}



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Ryan McKenna, © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally from https://github.com/ryan112358/private-pgm/blob/557c077708d3559212a8f65dff3eccd3fd244abb/src/mbi/callbacks.py
# Modified by the twinify Developers under the Apache 2.0 license
# Changed import statements

import time
import pandas as pd
import numpy as np


class CallBack:
    """ A CallBack is a function called after every iteration of an iterative optimization procedure
    It is useful for tracking loss and other metrics over time.
    """

    def __init__(self, engine, frequency=50):
        """ Initialize the callback objet
        :param engine: the FactoredInference object that is performing the optimization
        :param frequency: the number of iterations to perform before computing the callback function
        """
        self.engine = engine
        self.frequency = frequency
        self.calls = 0

    def run(self, marginals):
        pass

    def __call__(self, marginals):
        if self.calls == 0:
            self.start = time.time()
        if self.calls % self.frequency == 0:
            self.run(marginals)
        self.calls += 1


class Logger(CallBack):
    """ Logger is the default callback function.  It tracks the time, L1 loss, L2 loss, and
        optionally the total variation distance to the true query answers (when available).
        The last is for debugging purposes only - in practice the true answers can not  be observed.
    """

    def __init__(self, engine, true_answers=None, frequency=50):
        """ Initialize the callback objet
        :param engine: the FactoredInference object that is performing the optimization
        :param true_answers: a dictionary containing true answers to the measurement queries.
        :param frequency: the number of iterations to perform before computing the callback function
        """
        CallBack.__init__(self, engine, frequency)
        self.true_answers = true_answers
        self.idx = 0

    def setup(self):
        model = self.engine.model
        total = sum(model.domain.size(cl) for cl in model.cliques)
        print('Total clique size:', total, flush=True)
        # cl = max(model.cliques, key=lambda cl: model.domain.size(cl))
        # print('Maximal clique', cl, model.domain.size(cl), flush=True)
        cols = ['iteration', 'time', 'l1_loss', 'l2_loss', 'feasibility']
        if self.true_answers is not None:
            cols.append('variation')
        self.results = pd.DataFrame(columns=cols)
        print('\t\t'.join(cols), flush=True)

    def variational_distances(self, marginals):
        errors = []
        for Q, y, proj in self.true_answers:
            for cl in marginals:
                if set(proj) <= set(cl):
                    mu = marginals[cl].project(proj)
                    x = mu.values.flatten()
                    diff = Q.dot(x) - y
                    err = 0.5 * np.abs(diff).sum() / y.sum()
                    errors.append(err)
                    break
        return errors

    def primal_feasibility(self, mu):
        ans = 0
        count = 0
        for r in mu:
            for s in mu:
                if r == s: break
                d = tuple(set(r) & set(s))
                if len(d) > 0:
                    x = mu[r].project(d).datavector()
                    y = mu[s].project(d).datavector()
                    err = np.linalg.norm(x - y, 1)
                    ans += err
                    count += 1
        try:
            return ans / count
        except:
            return 0

    def run(self, marginals):
        if self.idx == 0:
            self.setup()

        t = time.time() - self.start
        l1_loss = self.engine._marginal_loss(marginals, metric='L1')[0]
        l2_loss = self.engine._marginal_loss(marginals, metric='L2')[0]
        feasibility = self.primal_feasibility(marginals)
        row = [self.calls, t, l1_loss, l2_loss, feasibility]
        if self.true_answers is not None:
            variational = np.mean(self.variational_distances(marginals))
            row.append(100 * variational)
        self.results.loc[self.idx] = row
        self.idx += 1

        print('\t\t'.join(['%.2f' % v for v in row]), flush=True)


# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Ryan McKenna, © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally from https://github.com/ryan112358/private-pgm/blob/557c077708d3559212a8f65dff3eccd3fd244abb/src/mbi/dataset.py
# Modified by the twinify Developers under the Apache 2.0 license
# Changed import statements

import numpy as np
import pandas as pd
import json
from twinify.napsu_mq.private_pgm.domain import Domain

class Dataset:
    def __init__(self, df, domain, weights=None):
        """ create a Dataset object
        :param df: a pandas dataframe
        :param domain: a domain object
        :param weight: weight for each row
        """
        assert set(domain.attrs) <= set(df.columns), 'data must contain domain attributes'
        assert weights is None or df.shape[0] == weights.size
        self.domain = domain
        self.df = df.loc[:, domain.attrs]
        self.weights = weights

    @staticmethod
    def synthetic(domain, N):
        """ Generate synthetic data conforming to the given domain
        :param domain: The domain object
        :param N: the number of individuals
        """
        arr = [np.random.randint(low=0, high=n, size=N) for n in domain.shape]
        values = np.array(arr).T
        df = pd.DataFrame(values, columns=domain.attrs)
        return Dataset(df, domain)

    @staticmethod
    def load(path, domain):
        """ Load data into a dataset object
        :param path: path to csv file
        :param domain: path to json file encoding the domain information
        """
        df = pd.read_csv(path)
        config = json.load(open(domain))
        domain = Domain(config.keys(), config.values())
        return Dataset(df, domain)

    def project(self, cols):
        """ project dataset onto a subset of columns """
        if type(cols) in [str, int]:
            cols = [cols]
        data = self.df.loc[:, cols]
        domain = self.domain.project(cols)
        return Dataset(data, domain, self.weights)

    def drop(self, cols):
        proj = [c for c in self.domain if c not in cols]
        return self.project(proj)

    @property
    def records(self):
        return self.df.shape[0]

    def datavector(self, flatten=True):
        """ return the database in vector-of-counts form """
        bins = [range(n + 1) for n in self.domain.shape]
        ans = np.histogramdd(self.df.values, bins, weights=self.weights)[0]
        return ans.flatten() if flatten else ans



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List, FrozenSet, Iterable
import pandas as pd
from functools import reduce

import twinify.napsu_mq
import twinify.cli.preprocessing_model as preprocessing_model

class ParsingError(Exception):
    pass


def _parse_model_file(model_file: str) -> Iterable[FrozenSet[str]]:
    with open(model_file, "r") as f:
        return [frozenset(line.split(",")) for line in  f.readlines()]


def load_cli_napsu(
        args: argparse.Namespace, unknown_args: Iterable[str], data_description: twinify.DataDescription
    ) -> twinify.dpvi.DPVIModel:

    column_feature_set = _parse_model_file(args.model_path)

    feature_names = reduce(set.union, column_feature_set, set())
    missing_features = feature_names.difference(data_description.columns)
    if missing_features:
        raise ParsingError(
            "The model specifies features that are not present in the data:\n{}".format(
                ", ".join(missing_features)
            )
        )

    drop_na = args.drop_na

    def preprocess_drop_na_fn(df: pd.DataFrame) -> pd.DataFrame:
        if drop_na:
            df = df.dropna()

        return df

    return preprocessing_model.PreprocessingModel(
        twinify.napsu_mq.NapsuMQModel(column_feature_set),
        preprocess_drop_na_fn
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Callable, Tuple

import jax
import jax.numpy as jnp
import numpyro
import numpyro.infer.util as nummcmc_util
from jax import random
import twinify.napsu_mq.maximum_entropy_model as mem
from twinify.napsu_mq.markov_network import MarkovNetwork


def run_numpyro_mcmc(
        rng: random.PRNGKey, suff_stat: jnp.ndarray, n: int, sigma_DP: float, max_ent_dist: MarkovNetwork,
        prior_mu: Union[float, jnp.ndarray] = 0, prior_sigma: float = 10, num_samples: int = 1000,
        num_warmup: int = 500, num_chains: int = 1, disable_progressbar: bool = False,
) -> numpyro.infer.MCMC:
    """Run MCMC inference (NUTS) with Numpyro on maximum entropy distribution with multivariate normal prior

    Args:
        rng (jax.random.PRNGKey): Jax random key for MCMC
        suff_stat (jax.numpy.ndarray): Noisy sufficient statistics array with DP noise added
        n (int): Number of datapoints
        sigma_DP (float): Noise standard deviation
        max_ent_dist (MarkovNetwork): Markov network representation of maximum entropy distribution
        prior_mu (float or jax.numpy.ndarray): Mean prior for multivariate normal distribution
        prior_sigma (float): Standard deviation prior for multivariate normal distribution
        num_samples (int): Number of samples for MCMC
        num_warmup (int): Number of warm-up (burn-in) samples for MCMC
        num_chains (int): Number of chains in MCMC
        disable_progressbar (bool): Disable progressbar for MCMC

    Returns:
        mcmc: Numpyro inference MCMC object
    """
    kernel = numpyro.infer.NUTS(model=mem.normal_prior_model_numpyro, max_tree_depth=12)
    mcmc = numpyro.infer.MCMC(
        kernel, num_warmup=num_warmup, num_samples=num_samples, num_chains=num_chains,
        progress_bar=not disable_progressbar, jit_model_args=False, chain_method="sequential"
    )
    mcmc.run(rng, suff_stat, n, sigma_DP, prior_mu, prior_sigma, max_ent_dist)
    return mcmc


def run_numpyro_mcmc_normalised(
        rng: random.PRNGKey, suff_stat: jnp.ndarray, n: int, sigma_DP: float, max_ent_dist: MarkovNetwork,
        laplace_approx: numpyro.distributions.MultivariateNormal, prior_sigma: float = 10,
        num_samples: int = 1000, num_warmup: int = 500, num_chains: int = 1, disable_progressbar: bool = False,
) -> Tuple[numpyro.infer.MCMC, Callable]:
    """Run MCMC inference (NUTS) with Numpyro on maximum entropy distribution with normalized multivariate normal prior

    Args:
        rng (jax.random.PRNGKey): Jax random key for MCMC
        suff_stat (jax.numpy.ndarray): Noisy sufficient statistics array with DP noise added
        n (int): Number of datapoints
        sigma_DP (float): Noise standard deviation
        max_ent_dist (MarkovNetwork): Markov network representation of maximum entropy distribution
        laplace_approx (numpyro.distributions.MultivariateNormal): Laplace approximation of multivariate normal for MCMC
        prior_sigma (float): Standard deviation prior for multivariate normal distribution
        num_samples (int): Number of samples for MCMC
        num_warmup (int): Number of warm-up (burn-in) samples for MCMC
        num_chains (int): Number of chains in MCMC
        disable_progressbar (bool): Disable progressbar for MCMC

    Returns:
        mcmc: Numpyro inference MCMC object
        backtransform: Function to transform posterior values back to non-normalized space
    """
    kernel = numpyro.infer.NUTS(model=mem.normal_prior_normalised_model_numpyro, max_tree_depth=12)
    mcmc = numpyro.infer.MCMC(
        kernel, num_warmup=num_warmup, num_samples=num_samples, num_chains=num_chains,
        progress_bar=not disable_progressbar, jit_model_args=False, chain_method="sequential"
    )

    mean_guess = laplace_approx.mean
    L_guess = jnp.linalg.cholesky(laplace_approx.covariance_matrix)
    mcmc.run(rng, suff_stat, n, sigma_DP, prior_sigma, max_ent_dist, mean_guess, L_guess)

    def backtransform(lambdas: jnp.ndarray) -> jnp.ndarray:
        return (L_guess @ lambdas.transpose()).transpose() + mean_guess

    return mcmc, backtransform


class ConvergenceException(Exception):
    """Convergence error in optimization process"""


def run_numpyro_laplace_approximation(
        rng: random.PRNGKey, suff_stat: jnp.ndarray, n: int, sigma_DP: float, max_ent_dist: MarkovNetwork,
        prior_mu: Union[float, jnp.ndarray] = 0, prior_sigma: float = 10, max_retries=5
) -> Tuple[numpyro.distributions.MultivariateNormal, bool]:
    """Run Laplace approximation on the maximum entropy distribution

    Args:
        rng (jax.random.PRNGKey): Jax random key for MCMC
        suff_stat (jax.numpy.ndarray): Noisy sufficient statistics array with DP noise added
        n (int): Number of datapoints
        sigma_DP (float): Noise standard deviation
        max_ent_dist (MarkovNetwork): Markov network representation of maximum entropy distribution
        prior_mu (float or jax.numpy.ndarray): Mean prior for multivariate normal distribution
        prior_sigma (float): Standard deviation prior for multivariate normal distribution
        max_retries (int): Times to retry the approximation

    Returns:
        laplace_approx: Laplace approximation for the maximum entropy distribution
        result.success: Boolean value if approximation was successful
    """

    key, *subkeys = random.split(rng, max_retries + 1)
    fail_count = 0

    for i in range(0, max_retries + 1):

        rng = subkeys[i]

        init_lambdas, potential_fn, t, mt = nummcmc_util.initialize_model(
            rng, mem.normal_prior_model_numpyro,
            model_args=(suff_stat, n, sigma_DP, prior_mu, prior_sigma, max_ent_dist)
        )

        lambdas = init_lambdas[0]["lambdas"]

        result = jax.scipy.optimize.minimize(lambda l: potential_fn({"lambdas": l}), lambdas, method="BFGS", tol=1e-2)
        if not result.success:
            fail_count += 1
        else:
            mean = result.x
            break

        if fail_count == max_retries:
            raise ConvergenceException(f"Minimize function failed to converge with {max_retries} retries")

    prec = jax.hessian(lambda l: potential_fn({"lambdas": l}))(mean)
    laplace_approx = numpyro.distributions.MultivariateNormal(loc=mean, precision_matrix=prec)
    return laplace_approx, result.success



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, List, Dict, Iterable, Tuple

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd

from twinify.napsu_mq.junction_tree import JunctionTree
from twinify.napsu_mq.undirected_graph import UndirectedGraph, greedy_ordering
from twinify.napsu_mq.marginal_query import FullMarginalQuerySet
from twinify.napsu_mq.log_factor import LogFactor
import functools


class MarkovNetwork:
    """Jax implementation of a Markov network representation for MED."""

    def __init__(self, domain: Dict, queries: FullMarginalQuerySet, elimination_order: Optional[Iterable] = None,
                 debug_checks: Optional[bool] = True):
        """Create the Markov network.
        Args:
            domain (dict): A dict containing the possible values for each variable.
            queries (FullMarginalQuerySet): The queries forming the sufficient statistic.
            elimination_order (list, optional): Elimination order for variable elimination. Defaults to using a greedy ordering.
            debug_checks (bool, optional): Whether to enable debug checks for factor computations. Defaults to True.
        """

        self.debug_checks = debug_checks
        self.domain = domain
        self.queries = queries
        self.d = len(self.domain.keys())
        variables_in_queries = set.union(*[set(feature_set) for feature_set in self.queries.feature_sets])
        self.variables_not_in_queries = set(domain.keys()).difference(variables_in_queries)

        self.graph = UndirectedGraph.from_clique_list(
            queries.feature_sets + [(var,) for var in self.variables_not_in_queries])
        if elimination_order is None:
            elimination_order = greedy_ordering(self.graph)
        self.elimination_order = elimination_order
        self.junction_tree = JunctionTree.from_variable_elimination(queries.feature_sets, self.elimination_order)
        self.junction_tree.remove_redundant_nodes()

        self.flat_queries = self.queries.flatten()
        self.variable_associations = self.flat_queries.get_variable_associations(use_features=True)
        self.suff_stat_d = len(self.flat_queries.queries)
        self.lambda_d = self.suff_stat_d

        self.suff_stat_mean = jax.jit(jax.grad(self.lambda0))
        self.suff_stat_cov = jax.jit(jax.hessian(self.lambda0))
        self.suff_stat_mean_bp = jax.jit(self.suff_stat_mean_bp)
        self.suff_stat_cov_bp = jax.jit(jax.jacrev(self.suff_stat_mean_bp))
        self.log_factor_class = LogFactor

    def suff_stat_mean_and_cov_bp(self, lambdas: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute sufficient statistic mean and covariance for given parameter values with belief propagation.
        Args:
            lambdas (numpy array): The parameter values.
        Returns:
            (numpy array, numpy array): (mean, covariance)
        """
        return self.suff_stat_mean_bp(lambdas), self.suff_stat_cov_bp(lambdas)

    def suff_stat_mean_and_cov(self, lambdas: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute sufficient statistic mean and covariance for given parameter values with.
        Args:
            lambdas (numpy array): The parameter values.
        Returns:
            (numpy array, numpy array): (mean, covariance)
        """
        return self.suff_stat_mean(lambdas), self.suff_stat_cov(lambdas)

    def marginal_distribution_logits(self, factors: Iterable[LogFactor], variables: List):
        to_eliminate = [var for var in self.elimination_order if var not in variables]
        result_factor = self.variable_elimination(factors, to_eliminate)
        proj_factor = result_factor.project(variables)
        proj_factor.ensure_batch_is_first_dim()
        return proj_factor.values

    def variable_elimination(self, factors: Iterable[LogFactor], variables_to_eliminate: Iterable) -> 'LogFactor':
        """Run variable elimination.
        Args:
            factors (list(LogFactor)): The factors used in variable elimination from self.compute_factors.
            variables_to_eliminate (list): The variables to eliminate.
        Returns:
            LogFactor: The result of variable elimination as a LogFactor.
        """
        for variable in variables_to_eliminate:
            factors = self.eliminate_var(factors, variable)

        return LogFactor.list_product(factors)

    @staticmethod
    def eliminate_var(factors: Iterable['LogFactor'], variable):
        factors_in_prod = [factor for factor in factors if variable in factor.scope]
        factors_not_in_prod = [factor for factor in factors if variable not in factor.scope]

        if len(factors_in_prod) == 0:
            return factors_not_in_prod
        log_prod_factor = LogFactor.list_product(factors_in_prod)

        summed_log_factor = log_prod_factor.marginalise(variable)
        factors_not_in_prod.append(summed_log_factor)
        return factors_not_in_prod

    def belief_propagation(self, factors: Iterable[LogFactor]) -> Dict:
        """Run belief propagation.
        Args:
            factors (list(LogFactor)): The factors used in belief propagation from self.compute_factors
        Returns:
            dict: Dict containing the LogFactor for each set factor scope in the Markov network.
        """

        self.junction_tree.init_factor_assignments()
        self.junction_tree.init_node_orders()

        for node in self.junction_tree.upward_order:
            node.reset()
            node.potential = LogFactor.list_product(
                factor for factor in factors if factor.scope in self.junction_tree.factors_in_node[node.variables]
            )
        for node in self.junction_tree.upward_order:
            if node.parent is not None:
                self.bp_message(node, node.parent)
        for node in self.junction_tree.downward_order:
            for child in node.children:
                self.bp_message(node, child)
        for node in self.junction_tree.downward_order:
            node.result = LogFactor.list_product([node.potential] + [message for _, message in node.messages])

        return {node.variables: node.result for node in self.junction_tree.downward_order}

    def bp_message(self, sender, receiver):
        product = LogFactor.list_product(
            [sender.potential] + [message for mes_sender, message in sender.messages if mes_sender is not receiver])
        edges = self.junction_tree.edges
        n1 = sender.variables
        n2 = receiver.variables
        separator = edges[(n1, n2)] if (n1, n2) in edges.keys() else edges[(n2, n1)]
        for variable in set(sender.variables).difference(set(separator)):
            product = product.marginalise(variable)
        receiver.messages.append((sender, product))

    @functools.partial(jax.jit, static_argnums=0)
    def lambda0(self, lambdas: jnp.ndarray) -> jnp.ndarray:
        factors = self.compute_factors(lambdas)
        result_factor = self.variable_elimination(factors, self.elimination_order)
        return result_factor.values

    def suff_stat_mean_bp(self, lambdas: jnp.ndarray) -> jnp.ndarray:
        factors = self.compute_factors(lambdas)
        result_factors = self.belief_propagation(factors)
        result: jnp.ndarray = jnp.zeros(self.suff_stat_d)
        for clique, indices in self.variable_associations.items():
            node_variables = self.junction_tree.node_for_factor[clique]
            factor = result_factors[node_variables]
            for variable in set(node_variables).difference(clique):
                factor = factor.marginalise(variable)
            result = result.at[jnp.array(indices)].set(factor.query(self.queries.queries[clique]))
        return result

    def sample(self, rng: jax.random.PRNGKey, lambdas: jnp.ndarray, n_sample: Optional[int] = 1) -> pd.DataFrame:
        """Sample the distribution with given parameter values.
        Args:
            lambdas (numpy array): The parameter values.
            n_sample (int, optional): The number of samples to generate. Defaults to 1.
        Returns:
            array (Pandas Dataframe): The generated samples.
        """
        n_cols = len(self.domain.keys())
        cols = self.domain.keys()
        data = np.zeros((n_sample, n_cols), dtype=np.int64)
        df = pd.DataFrame(data, columns=cols, dtype=int)
        order = self.elimination_order[::-1]
        batch_factors = [factor.add_batch_dim(n_sample) for factor in self.compute_factors(lambdas)]
        for variable in order:
            marginal = self.marginal_distribution_logits(batch_factors, [variable])
            rng, key = jax.random.split(rng)
            values = jax.random.categorical(key, marginal)
            batch_factors = [factor.batch_condition(variable, values) if variable in factor.scope else factor for factor
                             in batch_factors]
            df.loc[:, variable] = values

        return df

    def log_factor_vector(self, lambdas: jnp.ndarray, variables: Iterable) -> jnp.ndarray:
        vec = jnp.zeros(tuple(len(self.domain[var]) for var in variables))
        for query_ind in self.variable_associations[variables]:
            query_val = jnp.array(self.flat_queries.queries[query_ind].value)
            vec = vec.at[tuple(query_val)].set(lambdas[query_ind])
        return vec

    def compute_factors(self, lambdas: jnp.ndarray) -> List[LogFactor]:
        """Compute the LogFactor objects used by other methods for given parameters.
        Args:
            lambdas (numpy array): The parameters.
        Returns:
            list(LogFactor): The resulting factors.
        """
        return [
                   LogFactor(factor_scope, self.log_factor_vector(lambdas, factor_scope), self.debug_checks)
                   for factor_scope in self.variable_associations.keys()
               ] + [
                   LogFactor((variable,), jnp.zeros(len(self.domain[variable])), self.debug_checks)
                   for variable in self.variables_not_in_queries
               ]



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Ryan McKenna

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally from https://github.com/ryan112358/private-pgm/blob/557c077708d3559212a8f65dff3eccd3fd244abb/src/mbi/dataset.py

from functools import reduce


class Domain:
    def __init__(self, attrs, shape):
        """ Construct a Domain object

        :param attrs: a list or tuple of attribute names
        :param shape: a list or tuple of domain sizes for each attribute
        """
        assert len(attrs) == len(shape), 'dimensions must be equal'
        self.attrs = tuple(attrs)
        self.shape = tuple(shape)
        self.config = dict(zip(attrs, shape))

    @staticmethod
    def fromdict(config):
        """ Construct a Domain object from a dictionary of { attr : size } values """
        return Domain(config.keys(), config.values())

    def project(self, attrs):
        """ project the domain onto a subset of attributes

        :param attrs: the attributes to project onto
        :return: the projected Domain object
        """
        # return the projected domain
        if type(attrs) is str:
            attrs = [attrs]
        shape = tuple(self.config[a] for a in attrs)
        return Domain(attrs, shape)

    def marginalize(self, attrs):
        """ marginalize out some attributes from the domain (opposite of project)
        :param attrs: the attributes to marginalize out
        :return: the marginalized Domain object
        """
        proj = [a for a in self.attrs if not a in attrs]
        return self.project(proj)

    def axes(self, attrs):
        """ return the axes tuple for the given attributes
        :param attrs: the attributes
        :return: a tuple with the corresponding axes
        """
        return tuple(self.attrs.index(a) for a in attrs)

    def transpose(self, attrs):
        """ reorder the attributes in the domain object """
        return self.project(attrs)

    def invert(self, attrs):
        """ returns the attributes in the domain not in the list """
        return [a for a in self.attrs if a not in attrs]

    def merge(self, other):
        """ merge this domain object with another
        :param other: another Domain object
        :return: a new domain object covering the full domain
        Example:
        >>> D1 = Domain(['a','b'], [10,20])
        >>> D2 = Domain(['b','c'], [20,30])
        >>> D1.merge(D2)
        Domain(['a','b','c'], [10,20,30])
        """
        extra = other.marginalize(self.attrs)
        return Domain(self.attrs + extra.attrs, self.shape + extra.shape)

    def contains(self, other):
        """ determine if this domain contains another
        """
        return set(other.attrs) <= set(self.attrs)

    def size(self, attrs=None):
        """ return the total size of the domain """
        if attrs == None:
            return reduce(lambda x, y: x * y, self.shape, 1)
        return self.project(attrs).size()

    def sort(self, how='size'):
        """ return a new domain object, sorted by attribute size or attribute name """
        if how == 'size':
            attrs = sorted(self.attrs, key=self.size)
        elif how == 'name':
            attrs = sorted(self.attrs)
        return self.project(attrs)

    def canonical(self, attrs):
        """ return the canonical ordering of the attributes """
        return tuple(a for a in self.attrs if a in attrs)

    def __contains__(self, attr):
        return attr in self.attrs

    def __getitem__(self, a):
        """ return the size of an individual attribute
        :param a: the attribute
        """
        return self.config[a]

    def __iter__(self):
        """ iterator for the attributes in the domain """
        return self.attrs.__iter__()

    def __len__(self):
        return len(self.attrs)

    def __eq__(self, other):
        return self.attrs == other.attrs and self.shape == other.shape

    def __repr__(self):
        inner = ', '.join(['%s: %d' % x for x in zip(self.attrs, self.shape)])
        return 'Domain(%s)' % inner

    def __str__(self):
        return self.__repr__()


# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

# read version number
import importlib
spec = importlib.util.spec_from_file_location("version_module", "twinify/version.py")
version_module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(version_module)

setuptools.setup(
    name='twinify',
    version = version_module.VERSION,
    author="twinify Developers",
    author_email="lukas.m.prediger@aalto.fi",
    description="A software package for privacy-preserving generation of a synthetic twin to a given sensitive data set.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DPBayes/twinify",
    packages=setuptools.find_packages(include=['twinify', 'twinify.*']),
    python_requires='>=3.7',
    install_requires=[
        'pandas >= 1.3.4, < 2.0',
        'd3p >= 0.2.0, < 1.0',
        'tqdm >= 4.62, < 5.0',
        'numpy >= 1.21, < 2.0',
        'graphviz >= 0.20.1, < 1.0.0',
        'arviz >= 0.12.1, < 1.0.0',
        'networkx >= 2.6.0, < 3.0.0',
        'disjoint-set >= 0.7.0, < 1.0.0'
    ],
    extras_require = {
        'examples': [
            'xlrd < 2.0',
            'scikit-learn',
            'openpyxl'
        ],
        'dev': [
            'pytest'
        ],
        'compatible-dependencies': "d3p[compatible-dependencies]",
        'cpu': "d3p[cpu]",
        'cuda': "d3p[cuda]",
    },
    entry_points = {
        'console_scripts': [
            'twinify=twinify.cli.__main__:main',
            'twinify-tools=twinify.cli.tools.__main__:main'
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research"
     ],
)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Ryan McKenna

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally from https://github.com/ryan112358/private-pgm/blob/557c077708d3559212a8f65dff3eccd3fd244abb/src/mbi/junction_tree.py

import networkx as nx
import itertools
import numpy as np
from collections import OrderedDict


class JunctionTree:
    """ A JunctionTree is a transformation of a GraphicalModel into a tree structure.  It is used
        to find the maximal cliques in the graphical model, and for specifying the message passing
        order for belief propagation.  The JunctionTree is characterized by an elimination_order,
        which is chosen greedily by default, but may be passed in if desired.
    """

    def __init__(self, domain, cliques, elimination_order=None):
        self.cliques = [tuple(cl) for cl in cliques]
        self.domain = domain
        self.graph = self._make_graph()
        self.tree, self.order = self._make_tree(elimination_order)

    def maximal_cliques(self):
        """ return the list of maximal cliques in the model """
        # return list(self.tree.nodes())
        return list(nx.dfs_preorder_nodes(self.tree))

    def mp_order(self):
        """ return a valid message passing order """
        edges = set()
        messages = [(a, b) for a, b in self.tree.edges()] + [(b, a) for a, b in self.tree.edges()]
        for m1 in messages:
            for m2 in messages:
                if m1[1] == m2[0] and m1[0] != m2[1]:
                    edges.add((m1, m2))
        G = nx.DiGraph()
        G.add_nodes_from(messages)
        G.add_edges_from(edges)
        return list(nx.topological_sort(G))

    def separator_axes(self):
        return {(i, j): tuple(set(i) & set(j)) for i, j in self.mp_order()}

    def neighbors(self):
        return {i: set(self.tree.neighbors(i)) for i in self.maximal_cliques()}

    def _make_graph(self):
        G = nx.Graph()
        G.add_nodes_from(self.domain.attrs)
        for cl in self.cliques:
            G.add_edges_from(itertools.combinations(cl, 2))
        return G

    def _triangulated(self, order):
        edges = set()
        G = nx.Graph(self.graph)
        for node in order:
            tmp = set(itertools.combinations(G.neighbors(node), 2))
            edges |= tmp
            G.add_edges_from(tmp)
            G.remove_node(node)
        tri = nx.Graph(self.graph)
        tri.add_edges_from(edges)
        cliques = [tuple(c) for c in nx.find_cliques(tri)]
        cost = sum(self.domain.project(cl).size() for cl in cliques)
        return tri, cost

    def _greedy_order(self, stochastic=True):
        order = []
        domain, cliques = self.domain, self.cliques
        unmarked = list(domain.attrs)
        cliques = set(cliques)
        total_cost = 0
        for k in range(len(domain)):
            cost = OrderedDict()
            for a in unmarked:
                # all cliques that have a
                neighbors = list(filter(lambda cl: a in cl, cliques))
                # variables in this "super-clique"
                variables = tuple(set.union(set(), *map(set, neighbors)))
                # domain for the resulting factor
                newdom = domain.project(variables)
                # cost of removing a
                cost[a] = newdom.size()

            # find the best variable to eliminate
            if stochastic:
                choices = list(unmarked)
                costs = np.array([cost[a] for a in choices], dtype=float)
                probas = np.max(costs) - costs + 1
                probas /= probas.sum()
                i = np.random.choice(probas.size, p=probas)
                a = choices[i]
                # print(choices, probas)
            else:
                a = min(cost, key=lambda a: cost[a])

            # do some cleanup
            order.append(a)
            unmarked.remove(a)
            neighbors = list(filter(lambda cl: a in cl, cliques))
            variables = tuple(set.union(set(), *map(set, neighbors)) - {a})
            cliques -= set(neighbors)
            cliques.add(variables)
            total_cost += cost[a]

        return order, total_cost

    def _make_tree(self, order=None):
        if order is None:
            # orders = [self._greedy_order(stochastic=True) for _ in range(1000)]
            # orders.append(self._greedy_order(stochastic=False))
            # order = min(orders, key=lambda x: x[1])[0]
            order = self._greedy_order(stochastic=False)[0]
        elif type(order) is int:
            orders = [self._greedy_order(stochastic=False)] + [self._greedy_order(stochastic=True) for _ in
                                                               range(order)]
            order = min(orders, key=lambda x: x[1])[0]
        self.elimination_order = order
        tri, cost = self._triangulated(order)
        # cliques = [tuple(c) for c in nx.find_cliques(tri)]
        cliques = sorted([self.domain.canonical(c) for c in nx.find_cliques(tri)])
        complete = nx.Graph()
        complete.add_nodes_from(cliques)
        for c1, c2 in itertools.combinations(cliques, 2):
            wgt = len(set(c1) & set(c2))
            complete.add_edge(c1, c2, weight=-wgt)
        spanning = nx.minimum_spanning_tree(complete)
        return spanning, order


# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import numpyro
import numpyro.distributions as numdist
import jax.numpy as jnp
from twinify.napsu_mq.markov_network import MarkovNetwork


def normal_prior_model_numpyro(dp_suff_stat: jnp.ndarray, n: int, sigma_DP: float, prior_mu: Union[float, jnp.ndarray],
                               prior_sigma: float, med: MarkovNetwork):
    """NumPyro model for NAPSU-MQ with isotropic Gaussian prior.
    Args:
        dp_suff_stat (jax.ndarray): Noisy sufficient statistic.
        n (int): Number of datapoints.
        sigma_DP (float): Noise standard deviation.
        prior_mu (float or jax.ndarray): Prior mean.
        prior_sigma (float): Prior standard deviation.
        med (MarkovNetwork): An implementation of MED.
    """
    d = med.suff_stat_d
    lambda_d = med.lambda_d

    lambdas = numpyro.sample("lambdas", numdist.MultivariateNormal(jnp.ones(lambda_d) * prior_mu,
                                                                   covariance_matrix=jnp.eye(
                                                                       lambda_d) * prior_sigma ** 2))
    mean, cov = med.suff_stat_mean_and_cov(lambdas)
    return numpyro.sample("a_hat",
                          numdist.MultivariateNormal(n * mean, covariance_matrix=n * cov + jnp.eye(d) * sigma_DP ** 2),
                          obs=jnp.array(dp_suff_stat))


def normal_prior_normalised_model_numpyro(dp_suff_stat: jnp.ndarray, n: int, sigma_DP: float, prior_sigma: float,
                                          med: MarkovNetwork, mean_guess: jnp.ndarray, L_guess: jnp.ndarray):
    """NumPyro model for NAPSU-MQ with isotropic Gaussian prior and normalisation.
    Args:
        dp_suff_stat (jax.ndarray): Noisy sufficient statistic.
        n (int): Number of datapoints.
        sigma_DP (float): Noise standard deviation.
        prior_sigma (float): Prior standard deviation.
        med (MarkovNetwork): An implementation of MED.
        mean_guess (jax.ndarray): A guess for the posterior mean.
        L_guess (jax.ndarray): A guess for the Cholesky decomposition of posterior covariance.
    """
    d = med.suff_stat_d
    lambda_d = med.lambda_d

    prior_cov_inner = jnp.linalg.solve(L_guess, jnp.eye(lambda_d) * prior_sigma ** 2)
    norm_prior_cov = jnp.linalg.solve(L_guess, prior_cov_inner.transpose()).transpose()

    norm_prior_mu = jnp.linalg.solve(L_guess, -mean_guess.reshape((-1, 1))).flatten()

    norm_lambdas = numpyro.sample("norm_lambdas",
                                  numdist.MultivariateNormal(norm_prior_mu, covariance_matrix=norm_prior_cov))
    lambdas = L_guess @ norm_lambdas + mean_guess
    mean, cov = med.suff_stat_mean_and_cov(lambdas)
    return numpyro.sample("a_hat",
                          numdist.MultivariateNormal(n * mean, covariance_matrix=n * cov + jnp.eye(d) * sigma_DP ** 2),
                          obs=dp_suff_stat)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List, Tuple, Callable
import pandas as pd
import numpyro.infer.autoguide

import twinify
import twinify.dpvi
import twinify.dpvi.modelling.automodel as automodel
import twinify.cli.dpvi_numpyro_model_loading as model_loading
import twinify.cli.preprocessing_model as preprocessing_model

def _load_cli_dpvi_automodel(
        model_path: str,
        mixture_components: int,
        clipping_threshold: 1.,
        num_epochs: int,
        subsample_ratio: float,
        drop_na: bool
    ) -> twinify.dpvi.DPVIModel:

    with open(model_path, 'r') as f:
        model_str = "".join(f.readlines())

    features = automodel.parse_model(model_str)

    # build model
    model_fn = automodel.make_model(features, mixture_components)

    # build variational guide for optimization
    guide_fn = numpyro.infer.autoguide.AutoDiagonalNormal(model_fn)

    # postprocessing for automodel
    def preprocess_fn(df: pd.DataFrame) -> pd.DataFrame:

        # pick features from data according to model file
        feature_names = [feature.name for feature in features]
        missing_features = set(feature_names).difference(df.columns)
        if missing_features:
            raise automodel.ParsingError(
                "The model specifies features that are not present in the data:\n{}".format(
                    ", ".join(missing_features)
                )
            )

        df = df.loc[:, feature_names]

        if drop_na:
            df = df.dropna()

        for feature in features:
            df = feature.preprocess_data(df)

        print(f"After preprocessing the data has {df.shape[0]} entries with {df.shape[1]} features each.")

        return df

    postprocess_fn = automodel.postprocess_function_factory(features)

    model = twinify.dpvi.DPVIModel(
        model_fn, guide_fn,
        clipping_threshold,
        num_epochs,
        subsample_ratio
    )

    model = preprocessing_model.PreprocessingModel(
        model, preprocess_fn, postprocess_fn
    )

    return model


def load_cli_dpvi(
        args: argparse.Namespace, unknown_args: List[str], data_description: twinify.DataDescription
    ) -> twinify.dpvi.DPVIModel:

    if args.model_path[-3:] == '.py':
        try:
            model_fn, guide_fn, preprocess_fn, postprocess_fn = model_loading.load_custom_numpyro_model(
                args.model_path, args, unknown_args, data_description
            )

            model = twinify.dpvi.DPVIModel(
                model_fn, guide_fn,
                args.clipping_threshold,
                args.num_epochs,
                args.sampling_ratio
            )

            drop_na = args.drop_na

            def preprocess_drop_na_fn(df: pd.DataFrame) -> pd.DataFrame:
                df = preprocess_fn(df)

                if drop_na:
                    df = df.dropna()

                print(f"After preprocessing the data has {df.shape[0]} entries with {df.shape[1]} features each.")

                return df

            model = preprocessing_model.PreprocessingModel(
                model, preprocess_drop_na_fn, postprocess_fn
            )

            return model

        except (ModuleNotFoundError, FileNotFoundError) as e:
            raise automodel.ParsingError(
                "UNABLE TO READ THE MODEL FILE.:\n" + e.msg
            )

    else:
        print("Parsing model from txt file (was unable to read it as python module containing numpyro code)")
        return _load_cli_dpvi_automodel(
            args.model_path,
            args.k,
            args.clipping_threshold,
            args.num_epochs,
            args.sampling_ratio,
            args.drop_na
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Any, Union, BinaryIO
import numpy as np
import jax.numpy as jnp
import jax

ENDIANESS = 'big'
STR_ENCODING = 'utf8'
STRLEN_BYTES = 2
ITEMCOUNT_BYTES = 4
ARRLEN_BYTES = 8

def write_shape(shape_tuple: Tuple[int], writer: BinaryIO) -> None:
    num_vals = len(shape_tuple)
    writer.write(num_vals.to_bytes(ITEMCOUNT_BYTES, ENDIANESS))
    for val in shape_tuple:
        writer.write(val.to_bytes(ITEMCOUNT_BYTES, ENDIANESS))

def read_shape(reader: BinaryIO) -> Tuple[int]:
    num_vals = int.from_bytes(reader.read(ITEMCOUNT_BYTES), ENDIANESS, signed=False)
    return tuple(
        int.from_bytes(reader.read(ITEMCOUNT_BYTES), ENDIANESS, signed=False) for _ in range(num_vals)
    )

def write_dtype(dtype: np.dtype, writer: BinaryIO) -> None:
    dtype_str = str(dtype).encode(STR_ENCODING)
    dtype_str_len = len(dtype_str)
    writer.write(dtype_str_len.to_bytes(STRLEN_BYTES, ENDIANESS))
    writer.write(dtype_str)

def read_dtype(reader: BinaryIO) -> np.dtype:
    dtype_str_len = int.from_bytes(reader.read(STRLEN_BYTES), ENDIANESS, signed=False)
    dtype_str = reader.read(dtype_str_len).decode(STR_ENCODING)
    return np.dtype(dtype_str)

def write_array(arr: Union[np.ndarray, jnp.ndarray], writer: BinaryIO) -> None:
    shape = np.shape(arr)
    dtype = arr.dtype
    write_shape(shape, writer)
    write_dtype(dtype, writer)
    arr_bytes = arr.tobytes()
    arr_bytes_len = len(arr_bytes)
    writer.write(arr_bytes_len.to_bytes(ARRLEN_BYTES, ENDIANESS))
    writer.write(arr_bytes)

def read_array(reader: BinaryIO) -> np.ndarray:
    shape = read_shape(reader)
    dtype = read_dtype(reader)
    arr_bytes_len = int.from_bytes(reader.read(ARRLEN_BYTES), ENDIANESS, signed=False)
    arr_bytes = reader.read(arr_bytes_len)
    return np.frombuffer(arr_bytes, dtype=dtype).reshape(shape)

def write_params(params, writer: BinaryIO) -> jax.tree_util.PyTreeDef:
    flat_data, treedef = jax.tree_flatten(params)
    num_data_sites = len(flat_data)
    writer.write(num_data_sites.to_bytes(ITEMCOUNT_BYTES, ENDIANESS))

    for data_site in flat_data:
        if not isinstance(data_site, (np.ndarray, jnp.ndarray)):
            data_site = np.array(data_site)

        write_array(data_site, writer)

    return treedef

def read_params(reader: BinaryIO, treedef: jax.tree_util.PyTreeDef) -> Any:
    num_data_sites = int.from_bytes(reader.read(ITEMCOUNT_BYTES), ENDIANESS, signed=False)
    flat_data = [read_array(reader) for _ in range(num_data_sites)]
    return jax.tree_unflatten(treedef, flat_data)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

MAJOR_VERSION = 2
MINOR_VERSION = 0
PATCH_VERSION = 0
EXT_VERSION = ""

EXT_VERSION_SUFFIX = f"-{EXT_VERSION}" if len(EXT_VERSION) > 0 else ""

VERSION = f"{MAJOR_VERSION}.{MINOR_VERSION}.{PATCH_VERSION}{EXT_VERSION_SUFFIX}"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import pandas as pd
from typing import Union, Optional, Iterable, BinaryIO
import d3p.random

class InferenceModel(metaclass=abc.ABCMeta):
    """ A statistical model to generate privacy-preserving synthetic twins data sets from sensitive data. """

    @abc.abstractmethod
    def fit(self, data: pd.DataFrame, rng: d3p.random.PRNGState, epsilon: float, delta: float, **kwargs) -> 'InferenceResult':
        """ Compute the parameter posterior (approximation) for a given data set, hyperparameters and privacy bounds.

        Args:
            data: A `pandas.DataFrame` containing (sensitive) data.
            rng: A seeded state for the d3p.random secure random number generator.
            epsilon: Privacy bound ε.
            delta: Privacy bound δ.
            kwargs: Optional (model specific) hyperparameters.
        """
        pass


class InferenceResult(metaclass=abc.ABCMeta):
    """ A posterior parameter (approximation) resulting from privacy-preserving inference on a data set
    for a particular `InferenceModel`. """

    @abc.abstractmethod
    def generate(self,
            rng: d3p.random.PRNGState,
            num_parameter_samples: int,
            num_data_per_parameter_sample: int = 1,
            single_dataframe: bool = True) -> Union[Iterable[pd.DataFrame], pd.DataFrame]:
        """ Samples a number of samples from the parameter posterior (approximation) and generates the given number of
        data points per parameter samples.

        By default returns a single data frame samples from the posterior predictive distribution, i.e.,
        for each data records first a parameter value is drawn from the parameter posterior distribution, then
        the data record is sampled from the model conditioned on that parameter value. `num_parameter_samples`
        in this case determines the number of data records included in the returned data frame.

        This behavior can be customized to sample more than one data record per parameter sample by setting argument
        `num_data_per_parameter_sample` to a value larger than 1, in which case the total number of records
        returned is `num_parameter_samples * num_data_per_parameter_sample`.

        Setting `single_dataframe = False` causes the method to return an iterable collection of data frames,
        each of which contains all data records sampled for a single parameter samples, i.e., in this case
        this method returns `num_parameter_samples` data frames each of containing `num_data_per_parameter_sample`
        records.

        Each of the data frames "looks" like the original data this `InferenceResult` was obtained from,
        i.e., it has identical column names and categorical labels (if any).

        Args:
            - rng: A seeded state for the d3p.random secure random number generator.
            - num_parameter_samples: How often to sample from the parameter posterior approximation.
            - num_data_per_parameter_sample: How many data points to generate for each parameter sample.
            - single_dataframe: Whether to combine data samples into a single data frame or return separate data frames.
        """
        raise NotImplementedError(f"{self.__class__.__name__} does not implement generate_extended yet.")

    @classmethod
    def load(cls, file_path_or_io: Union[str, BinaryIO], **kwargs) -> 'InferenceResult':
        """ Loads an inference result from a file.

        The file can be specified either as a path or an opened file, i.e., both of the following
        are possible:
        ```InferenceResult.load('my_inference_result.out')```
        or
        ```
        with open('my_inference_result.out', 'rb') as f:
            InferenceResult.load(f)
        ```.

        Args:
            - file_path_or_io: The file path (as a string) or a `BinaryIO` instance representing the
                file from which to load.
            - kwargs: Optional (model specific) arguments for loading.

        If a `BinaryIO` instance is passed, the cursor position is advanced to after the data representing the
        inference result.

        Exceptions:
            raises an `InvalidFileFormatException` if the data in the file is not a valid representation of the
            inference result type represented by this class.

        Note for subclass implementation:
            Subclasses need only implement an override for `_load_from_io`.
        """
        if isinstance(file_path_or_io, str):
            with open(file_path_or_io, 'rb') as f:
                return cls._load_from_io(f, **kwargs)
        else:
            if not file_path_or_io.readable():
                raise ValueError("file_path_or_io is not a readable BinaryIO instance.")
            return cls._load_from_io(file_path_or_io, **kwargs)

    @classmethod
    @abc.abstractmethod
    def _load_from_io(cls, read_io: BinaryIO, **kwargs) -> 'InferenceResult':
        """ Internal implementation for `load` using a `BinaryIO` instance.

        Args:
            - read_io: A readable `BinaryIO` instance for reading binary data representing the inference result.
            - kwargs: Optional (model specific) arguments for loading.

        Note for subclass implementation:
            This method MUST raise an `InvalidFileFormatException` if the data stored in the file does not represent
            an inference result type implemented by this class (i.e., if the data cannot be loaded as an instance of
            this class).
        """
        raise NotImplementedError(f"{cls.__name__} does not implement load_from_io yet.")

    @classmethod
    def is_file_stored_result(cls, file_path_or_io: Union[str, BinaryIO], **kwargs) -> bool:
        """ Checks whether a file stores data representing the specific inference result type represented by this class.

        The file can be specified either as a path or an opened file, i.e., both of the following
        are possible:
        ```InferenceResult.is_file_stored_result('my_inference_result.out')```
        or
        ```
        with open('my_inference_result.out', 'wb') as f:
            InferenceResult.is_file_stored_result(f)
        ```.

        If a `BinaryIO` instance is passed, the cursor position will remain the same after this method returns.

        Args:
            - file_path_or_io: The file path (as a string) or a `BinaryIO` instance representing the
                file to check.
            - kwargs: Optional (model specific) arguments.

        Note for subclass implementation:
            Subclasses need only implement an override for `_is_file_stored_result_from_io`.
        """
        if isinstance(file_path_or_io, str):
            with open(file_path_or_io, 'rb') as f:
                return cls._is_file_stored_result_from_io(f)
        else:
            if not file_path_or_io.readable() or not file_path_or_io.seekable():
                raise ValueError("file_path_or_io is not a readable and seekable BinaryIO instance.")
            return cls._is_file_stored_result_from_io(file_path_or_io)

    @classmethod
    def _is_file_stored_result_from_io(cls, read_io: BinaryIO, **kwargs) -> bool:
        """ Internal implementation for `is_file_stored_result` using a `BinaryIO` instance.

        Args:
            - read_io: A readable and seekable `BinaryIO` instance for reading binary data representing the
                inference result.
            - kwargs: Optional (model specific) arguments.

        Note for subclass implementation:
            Default implementation tries to read the file and returns False if `_load_from_io` raises an
            `InvalidFileFormatException`. Subclasses should implement a more direct way of determining the type of
            inference result stored in the file (e.g. a type identifier in the beginning of the data). Implementation
            should additionally ensure that the implementation does not advance the cursor position of `read_io`.
        """
        try:
            cls._load_from_io(read_io, **kwargs)
            return True
        except InvalidFileFormatException:
            return False

    def store(self, file_path_or_io: Union[str, BinaryIO]) -> None:
        """ Writes the inference result to a file.

        The file can be specified either as a path or an opened file, i.e., both of the following
        are possible:
        ```inference_result.store('my_inference_result.out')```
        or
        ```
        with open('my_inference_result.out', 'wb') as f:
            inference_result.store(f)
        ```.

        Args:
            - file_path_or_io: The file path (as a string) or a `BinaryIO` instance representing the
                file to write data into.

        If a `BinaryIO` instance is passed, the cursor position is advanced to after the data representing the
        inference result.

        Note for subclass implementation:
            Subclasses only need to implement an override for `_store_to_io`.
        """
        if isinstance(file_path_or_io, str):
            with open(file_path_or_io, 'wb') as f:
                self._store_to_io(f)
        else:
            if not file_path_or_io.writable():
                raise ValueError("file_path_or_io is not a writable BinaryIO instance.")
            self._store_to_io(file_path_or_io)

    @abc.abstractmethod
    def _store_to_io(self, write_io: BinaryIO) -> None:
        """ Internal implementation for `store` using a `BinaryIO` instance.

        Args:
            - write_io: A writeable `BinaryIO` instance for writing binary data representing the inference result.

        Note for subclass implementation:
            The data written to the file should
            1) not rely overly much on pickling
            2) allow to identify the type of inference result written to the file
                (to identify the class with which to load it; see also `_is_file_stored_result_from_io`)
        """
        raise NotImplementedError(f"{self.__class__.__name__} does not implement store_to_io yet.")


class InvalidFileFormatException(Exception):

    def __init__(self, cls: type, msg: Optional[str] = None) -> None:
        msg = "Details:\n" + msg if msg is not None else ""
        super().__init__(f"The given file cannot be interpreted as {cls}.{msg}")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Ryan McKenna, © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally from https://github.com/ryan112358/private-pgm/blob/557c077708d3559212a8f65dff3eccd3fd244abb/src/mbi/__init__.py
# Modified by the twinify Developers under the Apache 2.0 license
# Modification contain changing the package folder and adding and changing import statements

from twinify.napsu_mq.private_pgm.clique_vector import CliqueVector
from twinify.napsu_mq.private_pgm.domain import Domain
from twinify.napsu_mq.private_pgm.dataset import Dataset
from twinify.napsu_mq.private_pgm.factor import Factor
from twinify.napsu_mq.private_pgm.graphical_model import GraphicalModel
from twinify.napsu_mq.private_pgm.inference import FactoredInference
from twinify.napsu_mq.private_pgm.junction_tree import JunctionTree
from twinify.napsu_mq.private_pgm.callbacks import Logger



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain, combinations
from typing import Iterable, List


def powerset(iterable: Iterable) -> List:
    """Create powerset from iterable
    powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    source: https://docs.python.org/3/library/itertools.html#itertools-recipes

    Args:
        iterable (Iterable): Iterable to create powerset from

    Returns:
        list (List): Powerset from iterable
    """
    s = list(iterable)
    return [set(t) for t in chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))]



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Missing value model used by twinify main script and available for modelling.
"""

import jax.numpy as jnp
import numpy as np
import jax
import typing

import numpyro.distributions as dists
from numpyro.distributions.transforms import biject_to


class _NAConstraint(dists.constraints.Constraint):
    """ Wraps a constraint to additionally allow NaN values.
    """

    def __init__(self, base_constraint: dists.constraints.Constraint) -> None:
        self.base_constraint = base_constraint
        super().__init__()

    def __call__(self, value: jnp.ndarray) -> jnp.ndarray:
        return jnp.isnan(value) | self.base_constraint(value)

    def feasible_like(self, prototype: jnp.ndarray) -> jnp.ndarray:
        return self.base_constraint.feasible_like(prototype)

    @property
    def is_discrete(self) -> bool:
        return self.base_constraint.is_discrete

    @property
    def event_dim(self) -> int:
        return self.base_constraint.event_dim


na_constraint = _NAConstraint
@biject_to.register(na_constraint)
def _transform_to_na(constraint):
    return biject_to(constraint.base_constraint)


class NAModel(dists.Distribution):
    """ Model decorator for missing values.

    Can be used to enhance any single feature distribution to handle missing
    values (represented by nan values (`np.nan`)) in a missing completely at random fashion.
    """

    arg_constraints = {'na_prob': dists.constraints.unit_interval}

    def __init__(self, base_dist, na_prob=0.5, *, validate_args=None) -> None:
        """
        Initialize a new NAModel instance.

        Args:
            base_dist (numpyro.distributions.Distribution): The feature distribution for existing data.
            na_prob (float): Probability that data is missing.
        """
        self._base_dist = base_dist
        self._na_prob = jnp.broadcast_to(na_prob, base_dist.batch_shape)
        assert self._base_dist.batch_shape == self._na_prob.shape
        super(NAModel, self).__init__(base_dist.batch_shape, base_dist.event_shape, validate_args=validate_args)

    @property
    def na_prob(self) -> float:
        return self._na_prob

    @property
    def base_distribution(self) -> dists.Distribution:
        """ The feature base distribution for existing data. """
        return self._base_dist

    @property
    def support(self) -> dists.constraints.Constraint:
        return na_constraint(self._base_dist.support)

    @property
    def has_enumerate_support(self) -> bool:
        return self._base_dist.has_enumerate_support

    def enumerate_support(self, expand=True):
        return self._base_dist.enumerate_support(expand)

    def log_prob(self, value: jnp.ndarray) -> jnp.ndarray:
        """
        Evaluates the log-probabilities for given data realizations (observed or missing)

        Args:
            values (array_like): The data.
        Returns:
            array_like containing probabilities for each given data point
        """
        if len(self.event_shape) > 0:
            reshaped_value = value.reshape(*(value.shape[:-len(self.event_shape)]), -1)
            nans = jnp.isnan(reshaped_value).sum(-1) == np.prod(self.event_shape)
        else:
            nans = jnp.isnan(value)
        nan_log_probs = dists.Bernoulli(probs = self._na_prob).log_prob(nans)

        feasible_vals = self.base_distribution.support.feasible_like(value)

        value_log_probs = self._base_dist.log_prob(
            jnp.where(
                jnp.expand_dims(nans, tuple(range(-len(self.event_shape), 0))),
                feasible_vals,
                value
            )
        )
        return nan_log_probs + jnp.where(nans, 0., value_log_probs)

    def sample(self, key :jax.random.KeyArray, sample_shape: typing.Tuple[int] = ()) -> jnp.ndarray:
        """
        Samples from the missing values model.

        Args:
            key (jax.random.PRNGKey): RNG key.
            sample_shape (tuple): How many samples to draw and how to arrange them
        Returns:
            array_like
        """
        return self.sample_with_intermediates(key, sample_shape)[0]

    def sample_with_intermediates(
            self,
            key :jax.random.KeyArray,
            sample_shape: typing.Tuple[int] = ()
        ) -> typing.Tuple[jnp.ndarray, typing.Tuple[typing.Any]]:

        nan_rng, values_rng = jax.random.split(key)
        nan_mask = dists.Bernoulli(probs = self._na_prob).sample(
            nan_rng, sample_shape=(sample_shape)
        )

        values = self.base_distribution.sample(values_rng, sample_shape=sample_shape)

        # expand trailing dims for nan_mask to match event_shape
        nan_mask = jnp.expand_dims(nan_mask, range(len(nan_mask.shape) - len(values.shape), 0))

        nans = jnp.where(nan_mask, jnp.nan, 0.)

        return values + nans, ()

    def __repr__(self) -> str:
        return f"NAModel({repr(self.base_distribution)}"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2022- twinify Developers and their Assignees

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import pickle
from typing import Optional, Union, Iterable, BinaryIO

import d3p.random
from twinify.base import InferenceModel, InferenceResult
from twinify import DataDescription
from twinify.cli.dpvi_numpyro_model_loading import TGuardedPreprocessFunction, TGuardedPostprocessFunction


class PostprocessingResult(InferenceResult):

    def __init__(self,
        base_result: InferenceResult,
        postprocessing_fn: TGuardedPostprocessFunction,
        data_description: DataDescription
    ) -> None:

        self._base_result = base_result
        self._postprocessing_fn = postprocessing_fn
        self._data_description = data_description
        super().__init__()

    @property
    def base(self) -> InferenceResult:
        return self._base_result

    @property
    def postprocess_fn(self) -> Optional[TGuardedPostprocessFunction]:
        return self._postprocessing_fn

    def postprocess(self, x: np.ndarray) -> pd.DataFrame:
        if self.postprocess_fn is not None:
            return self.postprocess_fn(x)
        return x

    def generate(self,
            rng: d3p.random.PRNGState,
            num_parameter_samples: int,
            num_data_per_parameter_sample: int = 1,
            single_dataframe: bool = True) -> Union[Iterable[pd.DataFrame], pd.DataFrame]:
        x = self._base_result.generate(rng, num_parameter_samples, num_data_per_parameter_sample, single_dataframe)
        x_post = self.postprocess(x)
        return x_post

    def _store_to_io(self, write_io: BinaryIO) -> None:
        base_type = type(self._base_result)
        pickle.dump(base_type, write_io)

        self._base_result._store_to_io(write_io)

        pickle.dump(self._data_description, write_io)


    @classmethod
    def _load_from_io(cls, read_io: BinaryIO, postprocessing_fn: TGuardedPostprocessFunction, **kwargs) -> InferenceResult:
        assert read_io.readable()

        base_type = pickle.load(read_io)
        base_result: InferenceResult = base_type._load_from_io(read_io, **kwargs)

        data_description: DataDescription = pickle.load(read_io)

        return PostprocessingResult(base_result, postprocessing_fn, data_description)


class PreprocessingModel(InferenceModel):

    def __init__(self,
        base_model: InferenceModel,
        preprocess_fn: Optional[TGuardedPreprocessFunction] = None,
        postprocess_fn: Optional[TGuardedPostprocessFunction] = None
    ) -> None:

        self._base_model = base_model
        self._preprocess_fn = preprocess_fn
        self._postprocess_fn = postprocess_fn

        super().__init__()

    @property
    def base(self) -> None:
        return self._base_model

    @property
    def preprocess_fn(self) -> Optional[TGuardedPostprocessFunction]:
        return self._preprocess_fn

    @property
    def postprocess_fn(self) -> Optional[TGuardedPostprocessFunction]:
        return self._postprocess_fn

    def preprocess(self, x: pd.DataFrame) -> pd.DataFrame:
        return self._preprocess_fn(x)

    def fit(self, data: pd.DataFrame, rng: d3p.random.PRNGState, epsilon: float, delta: float, **kwargs) -> InferenceResult:
        if self._preprocess_fn is not None:
            data = self._preprocess_fn(data.copy())

        result = self.base.fit(data, rng, epsilon, delta, **kwargs)

        if self._postprocess_fn is not None:
            data_description = DataDescription.from_dataframe(data)
            return PostprocessingResult(result, self._postprocess_fn, data_description)

        return result
